CHIPSET(0x29a2, i965,    "BW",      "Intel(R) 965G")
CHIPSET(0x2992, i965,    "BW",      "Intel(R) 965Q")
CHIPSET(0x2982, i965,    "BW",      "Intel(R) 965G")
CHIPSET(0x2972, i965,    "BW",      "Intel(R) 946GZ")
CHIPSET(0x2a02, i965,    "CL",      "Intel(R) 965GM")
CHIPSET(0x2a12, i965,    "CL",      "Intel(R) 965GME/GLE")

CHIPSET(0x2a42, g4x,     "CTG",     "Mobile Intel® GM45 Express Chipset")
CHIPSET(0x2e02, g4x,     "ELK",     "Intel(R) Integrated Graphics Device")
CHIPSET(0x2e12, g4x,     "ELK",     "Intel(R) Q45/Q43")
CHIPSET(0x2e22, g4x,     "ELK",     "Intel(R) G45/G43")
CHIPSET(0x2e32, g4x,     "ELK",     "Intel(R) G41")
CHIPSET(0x2e42, g4x,     "ELK",     "Intel(R) B43")
CHIPSET(0x2e92, g4x,     "ELK",     "Intel(R) B43")

CHIPSET(0x0042, ilk,     "ILK",     "Intel(R) HD Graphics")
CHIPSET(0x0046, ilk,     "ILK",     "Intel(R) HD Graphics")

CHIPSET(0x0102, snb_gt1, "SNB GT1", "Intel(R) HD Graphics 2000")
CHIPSET(0x0112, snb_gt2, "SNB GT2", "Intel(R) HD Graphics 3000")
CHIPSET(0x0122, snb_gt2, "SNB GT2", "Intel(R) HD Graphics 3000")
CHIPSET(0x0106, snb_gt1, "SNB GT1", "Intel(R) HD Graphics 2000")
CHIPSET(0x0116, snb_gt2, "SNB GT2", "Intel(R) HD Graphics 3000")
CHIPSET(0x0126, snb_gt2, "SNB GT2", "Intel(R) HD Graphics 3000")
CHIPSET(0x010a, snb_gt1, "SNB GT1", "Intel(R) HD Graphics 2000")

CHIPSET(0x0152, ivb_gt1, "IVB GT1", "Intel(R) HD Graphics 2500")
CHIPSET(0x0162, ivb_gt2, "IVB GT2", "Intel(R) HD Graphics 4000")
CHIPSET(0x0156, ivb_gt1, "IVB GT1", "Intel(R) HD Graphics 2500")
CHIPSET(0x0166, ivb_gt2, "IVB GT2", "Intel(R) HD Graphics 4000")
CHIPSET(0x015a, ivb_gt1, "IVB GT1", "Intel(R) HD Graphics")
CHIPSET(0x016a, ivb_gt2, "IVB GT2", "Intel(R) HD Graphics P4000")

CHIPSET(0x0402, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0412, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics 4600")
CHIPSET(0x0422, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x0406, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0416, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics 4600")
CHIPSET(0x0426, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x040a, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x041a, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics P4600/P4700")
CHIPSET(0x042a, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x040b, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x041b, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics")
CHIPSET(0x042b, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x040e, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x041e, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics 4400")
CHIPSET(0x042e, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x0c02, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0c12, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics")
CHIPSET(0x0c22, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x0c06, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0c16, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics")
CHIPSET(0x0c26, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x0c0a, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0c1a, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics")
CHIPSET(0x0c2a, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x0c0b, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0c1b, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics")
CHIPSET(0x0c2b, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x0c0e, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0c1e, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics")
CHIPSET(0x0c2e, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x0a02, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0a12, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics")
CHIPSET(0x0a22, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x0a06, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0a16, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics 4400")
CHIPSET(0x0a26, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics 5000")
CHIPSET(0x0a0a, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0a1a, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics")
CHIPSET(0x0a2a, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x0a0b, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0a1b, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics")
CHIPSET(0x0a2b, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x0a0e, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0a1e, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics 4200")
CHIPSET(0x0a2e, hsw_gt3, "HSW GT3", "Intel(R) Iris(R) Graphics 5100")
CHIPSET(0x0d02, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0d12, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics 4600")
CHIPSET(0x0d22, hsw_gt3, "HSW GT3", "Intel(R) Iris(R) Pro Graphics 5200")
CHIPSET(0x0d06, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0d16, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics")
CHIPSET(0x0d26, hsw_gt3, "HSW GT3", "Intel(R) Iris(R) Pro Graphics P5200")
CHIPSET(0x0d0a, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0d1a, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics")
CHIPSET(0x0d2a, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x0d0b, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0d1b, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics")
CHIPSET(0x0d2b, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")
CHIPSET(0x0d0e, hsw_gt1, "HSW GT1", "Intel(R) HD Graphics")
CHIPSET(0x0d1e, hsw_gt2, "HSW GT2", "Intel(R) HD Graphics")
CHIPSET(0x0d2e, hsw_gt3, "HSW GT3", "Intel(R) HD Graphics")

CHIPSET(0x0f31, byt,     "BYT",     "Intel(R) HD Graphics")
CHIPSET(0x0f32, byt,     "BYT",     "Intel(R) HD Graphics")
CHIPSET(0x0f33, byt,     "BYT",     "Intel(R) HD Graphics")
CHIPSET(0x0157, byt,     "BYT",     "Intel(R) HD Graphics")
CHIPSET(0x0155, byt,     "BYT",     "Intel(R) HD Graphics")

CHIPSET(0x22b0, chv,     "CHV",     "Intel(R) HD Graphics")
CHIPSET(0x22b1, chv,     "BSW",     "Intel(R) HD Graphics XXX") /* Overridden in brw_get_renderer_string */
CHIPSET(0x22b2, chv,     "CHV",     "Intel(R) HD Graphics")
CHIPSET(0x22b3, chv,     "CHV",     "Intel(R) HD Graphics")
