/*
 * Copyright © 2021 Ilia Mirkin <imirkin@alum.mit.edu>
 * SPDX-License-Identifier: MIT
 *
 * Authors:
 *    Ilia Mirkin <imirkin@alum.mit.edu>
 */

#ifndef FD4_COMPUTE_H_
#define FD4_COMPUTE_H_

#include "pipe/p_context.h"

void fd4_compute_init(struct pipe_context *pctx);

#endif /* FD4_COMPUTE_H_ */
